import cliutils, logging, json, options, os, osproc, sequtils, streams,
  strtabs, strutils, tables, uuids

import ./configuration
import nre except toSeq
from posix import link

type
  Workspace = ref object      ## Data needed by internal build process
    artifactsDir*: string     ## absolute path to the directory for this version
    buildDataDir*: string     ## absolute path to the global build data directory for this project
    buildRef*: string         ## git-style commit reference to the revision we are building
    dir*: string              ## absolute path to the working directory
    env*: StringTableRef      ## environment variables for all build processes
    openedFiles*: seq[File]   ## all files that we have opened that need to be closed
    outputHandler*: HandleProcMsgCB ## handler for process output
    project*: ProjectConfig   ## the project configuration
    projectDef*: ProjectDef   ## the StrawBoss project definition
    runRequest*: RunRequest   ## the RunRequest that initated the current build
    status*: BuildStatus      ## the current status of the build
    statusFile*: string       ## absolute path to the build status file
    step*: Step               ## the step we're building
    version*: string          ## project version as returned by versionCmd

  Worker* = object
    runId*: UUID
    process*: Process

proc sendMsg(h: HandleProcMsgCB, msg: TaintedString): void =
  h.sendMsg(msg, nil, "strawboss")

proc sendErrMsg(h: HandleProcMsgCB, msg: TaintedString): void =
  h.sendMsg(nil, msg, "strawboss")

proc resolveEnvVars(line: string, env: StringTableRef): string =
  result = line
  for found in line.findAll(re"\$\w+|\$\{[^}]+\}"):
    let key = if found[1] == '{': found[2..^2] else: found[1..^1]
    if env.hasKey(key): result = result.replace(found, env[key])

proc emitStatus(status: BuildStatus, statusFilePath: string,
                outputHandler: HandleProcMsgCB): BuildStatus =
  ## Emit a BuildStatus to the given file as a JSON object and to the given
  ## message handlers.
  if statusFilePath != nil: writeFile(statusFilePath, $status)
  if outputHandler != nil:
    outputHandler.sendMsg(status.state & ": " & status.details)
  result = status

proc publishStatus(wksp: Workspace, state, details: string) =
  ## Update the status for a Workspace and publish this status to the
  ## Workspace's status file and any output message handlers.
  let status = BuildStatus(
    runId: $wksp.runRequest.id, state: state, details: details)
  wksp.status = emitStatus(status, wksp.statusFile, wksp.outputHandler)

proc ensureProjectDirsExist(cfg: StrawBossConfig, p: ProjectDef): void =
  for subdir in ["configurations", "runs", "status", "artifacts"]:
    let fullPath = cfg.buildDataDir & "/" & p.name & "/" & subdir
    if not existsDir(fullPath):
      createDir(fullPath)

proc listVersions*(cfg: StrawBossConfig, project: ProjectDef): seq[string] =
  ## List the versions that have been built for a project.
  ensureProjectDirsExist(cfg, project)

  let versionFiles = toSeq(walkFiles(cfg.buildDataDir & "/" & project.name & "/configurations/*.json"))
  result = versionFiles.map(proc(s: string): string =
    let slashIdx = s.rfind('/')
    result = s[(slashIdx + 1)..^6])

proc listRuns*(cfg: StrawBossConfig, project: ProjectDef): seq[RunRequest] =
  ## List the runs that have been performed for a project.
  ensureProjectDirsExist(cfg, project)

  let runPaths = toSeq(walkFiles(cfg.buildDataDir & "/" & project.name & "/runs/*.request.json"))
  return runPaths.mapIt(parseRunRequest(parseFile(it)))

proc getCurrentProjectConfig*(cfg: StrawBossConfig, project: ProjectDef): Option[ProjectConfig] =
  let projCfgFile = "nope.json" # TODO
  if not existsFile(projCfgFile): result = none(ProjectConfig)
  else:
    try:
      let projectConfig: ProjectConfig = loadProjectConfig(projCfgFile) #ProjectConfig(name: "test")
      result = some(projectConfig)
    except: result = none(ProjectConfig)

proc setupProject(wksp: Workspace) =

  # Clone the project into the $temp directory
  let cloneResult = exec("git", ".",
    ["clone", wksp.projectDef.repo, wksp.dir],
    wksp.env, {poUsePath}, wksp.outputHandler)

  if cloneResult != 0:
    raiseEx "unable to clone repo for '" & wksp.projectDef.name & "'"

  # Checkout the requested ref
  let checkoutResult = exec("git", wksp.dir,
    ["checkout", wksp.buildRef],
    wksp.env, {poUsePath}, wksp.outputHandler)

  if checkoutResult != 0:
    raiseEx "unable to checkout ref " & wksp.buildRef &
            " for '" & wksp.projectDef.name & "'"

  # Find the strawboss project configuration
  let projCfgFile = wksp.dir & wksp.projectDef.cfgFilePath
  if not existsFile(projCfgFile):
    raiseEx "Cannot find strawboss project configuration in the project " &
      "repo (expected at '" & wksp.projectDef.cfgFilePath & "')."

  wksp.project = loadProjectConfig(projCfgFile)

  # Merge in the project-defined env vars
  for k, v in wksp.projectDef.envVars: wksp.env[k] = v

  # Get the build version
  let versionResult = execWithOutput(
    wksp.project.versionCmd,    # command
    wksp.dir,                   # working dir
    [],                         # args
    wksp.env,                   # environment
    {poUsePath, poEvalCommand}) # options

  if versionResult.exitCode != 0:
    raiseEx "Version command (" & wksp.project.versionCmd &
            ") returned non-zero exit code."

  wksp.version = versionResult.output.strip
  wksp.env["VERSION"] = wksp.version

proc runStep*(wksp: Workspace, step: Step) =

  ## Lower-level method to execute a given step within the context of a project
  ## workspace that is setup and configured. May be called recursively to
  ## satisfy step dependencies.

  let SB_EXPECTED_VARS = ["VERSION"]

  wksp.publishStatus("running",
    "running '" & step.name & "' for version " & wksp.version &
    " from " & wksp.buildRef)

  # Ensure all expected environment variables are present.
  for k in (step.expectedEnv & @SB_EXPECTED_VARS):
    if not wksp.env.hasKey(k):
      raiseEx "step " & step.name & " failed: missing required env variable: " & k

  # Ensure that artifacts in steps we depend on are present
  # TODO: detect circular-references in dependency trees.
  for dep in step.depends:
    if not wksp.project.steps.hasKey(dep):
      raiseEx step.name & " depends on " & dep &
              " but there is no step named " & dep
    let depStep = wksp.project.steps[dep]

    # Run that step (may get skipped)
    runStep(wksp, depStep)

    # Add the artifacts directory for the dependent step to our env so that
    # further steps can reference it via $<stepname>_DIR
    wksp.env[depStep.name & "_DIR"] = wksp.buildDataDir & "/artifacts/" &
      "/" & dep & "/" & wksp.version

  # Run the step command, piping in cmdInput
  wksp.outputHandler.sendMsg step.name & ": starting stepCmd: " & step.stepCmd
  let cmdProc = startProcess(step.stepCmd,
    wksp.dir & step.workingDir, [], wksp.env, {poUsePath, poEvalCommand})

  let cmdInStream = inputStream(cmdProc)

  # Replace env variables in step cmdInput as we pipe it in
  for line in step.cmdInput: cmdInStream.writeLine(line.resolveEnvVars(wksp.env))
  cmdInStream.flush()
  cmdInStream.close()
  
  let cmdResult = waitFor(cmdProc, wksp.outputHandler, step.stepCmd)

  if cmdResult != 0:
    raiseEx "step " & step.name & " failed: step command returned non-zero exit code"

  # Gather the output artifacts (if we have any)
  wksp.outputHandler.sendMsg "artifacts: " & $step.artifacts
  if step.artifacts.len > 0:
    for a in step.artifacts:
      let artifactPath = a.resolveEnvVars(wksp.env)
      let artifactName = artifactPath[(artifactPath.rfind("/")+1)..^1]
      try:
        wksp.outputHandler.sendMsg "copy " & wksp.dir &
          step.workingDir & "/" & artifactPath & " -> " &
          wksp.artifactsDir & "/" & artifactName

        copyFile(wksp.dir & step.workingDir & "/" & artifactPath,
                 wksp.artifactsDir & "/" & artifactName)
      except:
        raiseEx "step " & step.name & " failed: unable to copy artifact " &
          artifactPath & ":\n" & getCurrentExceptionMsg()

  wksp.publishStatus("complete", "")

proc initiateRun*(cfg: StrawBossConfig, req: RunRequest,
              outputHandler: HandleProcMsgCB = nil): BuildStatus =

  ## Execute a RunReuest given the StrawBoss configuration. This is the main
  ## entrypoint to  running a build step.

  result = BuildStatus(
    runId: $req.id,
    state: "setup",
    details: "initializing build workspace")
  discard emitStatus(result, nil, outputHandler)

  var wksp: Workspace

  try:
    # Find the project definition
    let projectDef = cfg.findProject(req.projectName)

    # Make sure the build data directories for this project exist.
    ensureProjectDirsExist(cfg, projectDef)

    # Update our run status
    let runDir = cfg.buildDataDir & "/" & projectDef.name & "/runs"
    discard emitStatus(result, runDir & "/" & $req.id & ".status.json", nil)

    # Read in the existing system environment
    var env = loadEnv()
    env["GIT_DIR"] = ".git"

    # Make sure we have a workspace directory
    assert req.workspaceDir.isAbsolute
    if not existsDir(req.workspaceDir): createDir(req.workspaceDir)

    # Setup our STDOUT and STDERR files
    let stdoutFile = open(runDir & "/" & $req.id & ".stdout.log", fmWrite)
    let stderrFile = open(runDir & "/" & $req.id & ".stderr.log", fmWrite)

    let logFilesOH = makeProcMsgHandler(stdoutFile, stderrFile)

    wksp = Workspace(
      artifactsDir: nil,
      buildDataDir: cfg.buildDataDir & "/" & projectDef.name,
      buildRef:
        if req.buildRef != nil and req.buildRef.len > 0: req.buildRef
        else: projectDef.defaultBranch,
      dir: req.workspaceDir,
      env: env,
      openedFiles: @[stdoutFile, stderrFile],
      outputHandler: combineProcMsgHandlers(outputHandler, logFilesOH),
      project: ProjectConfig(),
      projectDef: projectDef,
      runRequest: req,
      status: result,
      statusFile: runDir & "/" & $req.id & ".status.json",
      step: Step(),
      version: nil)

  except:
    when not defined(release): echo getCurrentException().getStackTrace()
    result = BuildStatus(runId: $req.id, state: "failed",
                         details: getCurrentExceptionMsg())
    try: discard emitStatus(result, nil, outputHandler)
    except: discard ""
    return

  try:
    # Clone the repo and setup the working environment
    wksp.publishStatus("setup",
      "cloning project repo and preparing to run '" & req.stepName & "'")
    wksp.setupProject()

    # Update our cache of project configurations.
    # TODO: what happens if this fails?
    copyFile(
      wksp.dir & "/" & wksp.projectDef.cfgFilePath,
      wksp.buildDataDir & "/configurations/" & wksp.version & ".json")

    # Find the requested step
    if not wksp.project.steps.hasKey(req.stepName):
      raiseEx "no step name '" & req.stepName & "' for " & req.projectName
    var step = wksp.project.steps[req.stepName]

    if req.forceRebuild: step.dontSkip = true

    wksp.artifactsDir = wksp.buildDataDir & "/artifacts/" &
                        step.name & "/" & wksp.version

    # Have we tried to build this before and are we caching the results?
    let statusFilePath = wksp.buildDataDir & "/status/" & wksp.version & ".json"
    if existsFile(statusFilePath) and not step.dontSkip:
      let prevStatus = loadBuildStatus(statusFilePath)

      # If we succeeded last time, no need to rebuild
      if prevStatus.state == "complete":
        wksp.outputHandler.sendMsg(
          "Skipping step '" & step.name & "' for version '" &
          wksp.version & "': already completed.")
        return prevStatus
      else:
        wksp.outputHandler.sendMsg(
          "Rebuilding failed step '" & step.name & "' for version '" &
          wksp.version & "'.")

    if not existsDir(wksp.artifactsDir): createDir(wksp.artifactsDir)

    runStep(wksp, step)

    # Record the results of this build as the status for this version. 
    writeFile(wksp.buildDataDir & "/status/" & wksp.version & ".json", $wksp.status)

    result = wksp.status

  except:
    when not defined(release): echo getCurrentException().getStackTrace()
    let msg = getCurrentExceptionMsg()
    try:
      wksp.publishStatus("failed", msg)
      result = wksp.status
    except:
      result = BuildStatus(runId: $req.id, state: "failed", details: msg)
      try: discard emitStatus(result, nil, outputHandler)
      except: discard ""

  finally:
    if wksp != nil:
      for f in wksp.openedFiles:
        try: close(f)
        except: discard ""

proc spawnWorker*(cfg: StrawBossConfig, req: RunRequest): Worker =
 
  # Find the project definition (will throw appropriate exceptions)
  let projectDef = cfg.findProject(req.projectName)
  let runDir = cfg.buildDataDir & "/" & projectDef.name & "/runs"
  let reqFile = runDir & "/" & $req.id & ".request.json"
  let statusFile = runDir & "/" & $req.id & ".status.json"

  try:
    # Make sure the build data directories for this project exist.
    ensureProjectDirsExist(cfg, projectDef)

    # Save the run request
    writeFile(reqFile, $req)

    # Write the initial build status (queued).
    let queuedStatus = BuildStatus(
      runId: $req.id,
      state: "queued",
      details: "request queued for execution")
    writeFile(statusFile, $queuedStatus)

    var args = @["run", reqFile]
    debug "Launching worker: " & cfg.pathToExe & " " & args.join(" ")
    result = Worker(
      runId: req.id,
      process: startProcess(cfg.pathToExe, ".", args, loadEnv(), {poUsePath}))

  except:
    let exMsg = "run request rejected: " & getCurrentExceptionMsg()
    raiseEx exMsg
    try:
      writeFile(statusFile,
        $(BuildStatus(runId: $req.id, state: "rejected", details: exMsg)))
    except: discard ""
